@@ -23,12 +23,11 @@
  ***************************************************************************/
 #include "tool_setup.h"
 
-#include "strcase.h"
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 
-#include "curlx.h"
+#include <curlx.h>
 
 #include "tool_cfgable.h"
 #include "tool_doswin.h"
@@ -38,7 +37,7 @@
 #include "tool_operate.h"
 #include "tool_libinfo.h"
 
-#include "memdebug.h" /* keep this as LAST include */
+#include <memdebug.h> /* keep this as LAST include */
 
 static char *parse_filename(const char *ptr, size_t len);
 
@@ -352,7 +351,7 @@ static char *parse_filename(const char *ptr, size_t len)
   if(q) {
     p = q + 1;
     if(!*p) {
-      curlx_safefree(copy);
+      tool_safefree(copy);
       return NULL;
     }
   }
@@ -364,7 +363,7 @@ static char *parse_filename(const char *ptr, size_t len)
   if(q) {
     p = q + 1;
     if(!*p) {
-      curlx_safefree(copy);
+      tool_safefree(copy);
       return NULL;
     }
   }
@@ -385,7 +384,7 @@ static char *parse_filename(const char *ptr, size_t len)
   {
     char *sanitized;
     SANITIZEcode sc = sanitize_file_name(&sanitized, copy, 0);
-    curlx_safefree(copy);
+    tool_safefree(copy);
     if(sc)
       return NULL;
     copy = sanitized;
